#!/usr/bin/env python3
# coding: utf-8
# pylint: disable=import-outside-toplevel
import json
from typing import List

from flask import current_app

from sapporo.model import (Log, RunId, RunListResponse, RunLog, RunStatus,
                           ServiceInfo, Workflow)
from sapporo.trs import get_wfs


def generate_service_info() -> ServiceInfo:
    from sapporo.run import count_system_state
    with current_app.config["SERVICE_INFO"].open(mode="r", encoding="utf-8") as f:
        service_info: ServiceInfo = json.load(f)

    service_info["supported_wes_versions"] = ["sapporo-wes-1.0.1"]
    service_info["system_state_counts"] = count_system_state()
    service_info["tags"]["get_runs"] = current_app.config["GET_RUNS"]
    service_info["tags"]["workflow_attachment"] = current_app.config["WORKFLOW_ATTACHMENT"]
    service_info["tags"]["registered_only_mode"] = current_app.config["REGISTERED_ONLY_MODE"]

    return service_info


def generate_executable_workflows() -> List[Workflow]:
    with current_app.config["EXECUTABLE_WORKFLOWS"].open(mode="r", encoding="utf-8") as f:
        data = json.load(f)
        executable_workflows: List[Workflow] = data["workflow"]
        trs_endpoints = data["trs_endpoint"]
    for endpoint in trs_endpoints:
        trs_wfs = get_wfs(endpoint)
        executable_workflows.extend(trs_wfs)

    return executable_workflows


def generate_run_status(run_id: str) -> RunStatus:
    from sapporo.run import read_state
    return {
        "run_id": run_id,
        "state": read_state(run_id)
    }


def generate_run_list() -> RunListResponse:
    from sapporo.run import glob_all_run_ids
    all_run_ids = glob_all_run_ids()
    runs: List[RunStatus] = [generate_run_status(
        run_id) for run_id in all_run_ids]

    return {
        "runs": runs,
        "next_page_token": ""
    }


def generate_run_log(run_id: str) -> RunLog:
    from sapporo.run import read_file, read_state
    run_log: RunLog = {
        "run_id": run_id,
        "request": read_file(run_id, "run_request"),
        "state": read_state(run_id),
        "run_log": generate_log(run_id),
        "task_logs": read_file(run_id, "task_logs"),
        "outputs": read_file(run_id, "outputs")
    }

    return run_log


def generate_log(run_id: str) -> Log:
    from sapporo.run import read_file
    log: Log = {
        "name": read_file(run_id, "run_request")["workflow_name"],
        "cmd": read_file(run_id, "cmd"),
        "start_time": read_file(run_id, "start_time"),
        "end_time": read_file(run_id, "end_time"),
        "stdout": read_file(run_id, "stdout"),
        "stderr": read_file(run_id, "stderr"),
        "exit_code": read_file(run_id, "exit_code")
    }

    return log


def generate_run_id(run_id: str) -> RunId:
    return {
        "run_id": run_id
    }



#!/usr/bin/env python3
# coding: utf-8
import json
from typing import List, Optional, Tuple, cast

from flask import abort, current_app, request
from werkzeug.datastructures import FileStorage

from sapporo.model import AttachedFile, ParseRequest, RunRequest, WorkflowTypes
from sapporo.model.factory import (generate_executable_workflows,
                                   generate_service_info)
from sapporo.parser import parse_workflows
from sapporo.run import (glob_all_run_ids, resolve_content_path,
                         sapporo_endpoint, secure_filepath)

EXPECTED_PARSING_TYPES = [
    "workflow_type",
    "workflow_type_version",
    "inputs",
    "make_template",
]


def validate_post_parse_workflows() -> ParseRequest:
    parse_request: ParseRequest = {
        "workflow_content": request.form.get("workflow_content", None),
        "workflow_location": request.form.get("workflow_location", None),
        "types_of_parsing": request.form.getlist("types_of_parsing") or ["workflow_type", "workflow_type_version"],
    }
    if parse_request["types_of_parsing"] is not None:
        for _type in parse_request["types_of_parsing"]:
            if _type not in EXPECTED_PARSING_TYPES:
                abort(400, f"The type of parsing `{_type}` is not supported.")
    if parse_request["workflow_content"] is None and parse_request["workflow_location"] is None:
        abort(400, "Either `workflow_content` or `workflow_location` must be specified.")
    if parse_request["workflow_content"] is not None and parse_request["workflow_location"] is not None:
        abort(400, "Either `workflow_content` or `workflow_location` must be specified, but not both.")

    return parse_request


def validate_get_runs() -> None:
    if current_app.config["GET_RUNS"] is False:
        abort(403, "This endpoint `GET /runs` is unavailable because the service provider didn't allow the request to this endpoint when sapporo was started.")


def validate_run_request(run_id: str) -> RunRequest:
    # Fields: do not validation
    wf_params = request.form.get("workflow_params", None)
    wf_engine_params = request.form.get("workflow_engine_parameters", None)
    tags = request.form.get("tags", None)

    # Fields: do validation, but not related to the registered mode
    wf_engine_name = request.form.get("workflow_engine_name", None)
    wf_attachment_str = request.form.get("workflow_attachment", None)
    wf_attachment_files = request.files.getlist("workflow_attachment")
    wf_engine_name = validate_wf_engine_name(wf_engine_name)
    wf_attachment = validate_wf_attachment(run_id, wf_attachment_str, wf_attachment_files)

    # Fields: do validation, related to the registered mode
    wf_url = request.form.get("workflow_url", None)
    wf_name = request.form.get("workflow_name", None)
    wf_type = request.form.get("workflow_type", None)
    wf_type_version = request.form.get("workflow_type_version", None)

    if wf_name is not None:  # registered mode request
        wf_url, wf_type, wf_type_version, wf_attachment = \
            validate_wf_docs_with_registered_wf(wf_name, wf_attachment)
    else:  # normal mode request
        validate_registered_only_mode()
        wf_type, wf_type_version = \
            validate_wf_docs_with_no_registered_wf(wf_url, wf_type, wf_type_version)
        wf_url = cast(str, wf_url)

    validate_wf_type(wf_type, wf_type_version)
    wf_type = cast(WorkflowTypes, wf_type)

    # Meta characters validation
    validate_meta_characters("workflow_url", wf_url)
    if wf_engine_params is not None:
        try:
            wf_engine_params_obj = json.loads(wf_engine_params)
            if isinstance(wf_engine_params_obj, list):
                for val in wf_engine_params_obj:
                    validate_meta_characters("workflow_engine_parameters", val)
            elif isinstance(wf_engine_params_obj, dict):
                for key, val in wf_engine_params_obj.items():
                    validate_meta_characters("workflow_engine_parameters", key)
                    validate_meta_characters("workflow_engine_parameters", val)
            else:
                abort(400, "The `workflow_engine_parameters` must be a list or a dict.")
        except json.JSONDecodeError:
            abort(400, "The `workflow_engine_parameters` is not valid JSON.")

    return {
        "workflow_params": wf_params,
        "workflow_type": wf_type,
        "workflow_type_version": wf_type_version,
        "tags": tags,
        "workflow_engine_name": wf_engine_name,
        "workflow_engine_parameters": wf_engine_params,
        "workflow_url": wf_url,
        "workflow_name": wf_name,
        "workflow_attachment": json.dumps(wf_attachment),
    }


def validate_wf_engine_name(wf_engine_name: Optional[str]) -> str:
    if wf_engine_name is None:
        abort(400, "The `workflow_engine_name` is required.")
    service_info = generate_service_info()
    wf_engines_names = list(service_info["workflow_engine_versions"].keys())
    if wf_engine_name not in wf_engines_names:
        abort(
            400, f"The workflow engine name `{wf_engine_name}` is not supported.")

    return wf_engine_name


def validate_registered_only_mode() -> None:
    if current_app.config["REGISTERED_ONLY_MODE"]:
        abort(403, "Currently, the sapporo-service is running in `registered-only-mode`. "
                   "Therefore, the endpoint `POST /runs` is unavailable to specify `workflow_url`.")


def validate_wf_attachment(run_id: str, wf_attachment_str: Optional[str], wf_attachment_files: List[FileStorage]) -> List[AttachedFile]:
    wf_attachment: List[AttachedFile] = []
    if current_app.config["WORKFLOW_ATTACHMENT"]:
        if wf_attachment_str is not None:
            try:
                wf_attachment_obj = json.loads(wf_attachment_str)
                try:
                    for attached_file in wf_attachment_obj:
                        wf_attachment.append({
                            "file_name": str(attached_file["file_name"]),
                            "file_url": str(attached_file["file_url"]),
                        })
                except Exception:
                    abort(
                        400, "The `workflow_attachment` must be a list of `AttachedFile`.")
            except json.JSONDecodeError:
                abort(400, "The `workflow_attachment` is invalid.")
        if len(wf_attachment_files):
            exe_dir = resolve_content_path(run_id, "exe_dir")
            endpoint = sapporo_endpoint()
            base_remote_url = f"{endpoint}/runs/{run_id}/data/"
            for file in wf_attachment_files:
                if file.filename is None:
                    continue
                file_name = secure_filepath(file.filename)
                file_path = exe_dir.joinpath(file_name).resolve()
                wf_attachment.append({
                    "file_name": str(file_name),
                    "file_url": base_remote_url + str(file_path.relative_to(exe_dir.parent))
                })
    else:
        if wf_attachment_str is not None or len(wf_attachment_files) != 0:
            abort(400, "The `workflow_attachment` is invalid because the sapporo-service is running in `disable-workflow-attachment`.")

    return wf_attachment


def validate_wf_docs_with_registered_wf(wf_name: str, wf_attachment: List[AttachedFile]) -> Tuple[str, str, str, List[AttachedFile]]:
    executable_wfs = generate_executable_workflows()
    wf_names = [wf["workflow_name"] for wf in executable_wfs]
    if wf_name not in wf_names:
        abort(400, f"The workflow name `{wf_name}` does not exist.")
    wf = executable_wfs[wf_names.index(wf_name)]
    wf_url = wf["workflow_url"]
    wf_type = wf["workflow_type"]
    wf_type_version = wf["workflow_type_version"]
    wf_attachment_names = [f["file_name"] for f in wf_attachment]
    for file in wf["workflow_attachment"]:
        if file["file_name"] not in wf_attachment_names:
            wf_attachment.append(file)

    return wf_url, wf_type, wf_type_version, wf_attachment


def validate_wf_docs_with_no_registered_wf(wf_url: Optional[str], wf_type: Optional[str], wf_type_version: Optional[str]) -> Tuple[str, str]:
    if wf_url is None:
        abort(400, "The `workflow_url` is required.")
    if wf_type is None or wf_type_version is None:
        try:
            parse_result = parse_workflows({
                "workflow_content": None,
                "workflow_location": wf_url,
                "types_of_parsing": ["workflow_type", "workflow_type_version"],
            })
            wf_type = wf_type or parse_result["workflow_type"]
            wf_type_version = wf_type_version or parse_result["workflow_type_version"]
        except Exception:
            abort(400, "The `workflow_type` and `workflow_type_version` are required.")

    return wf_type, wf_type_version  # type: ignore


def validate_wf_type(wf_type: str, wf_type_version: str) -> None:
    service_info = generate_service_info()
    wf_type_versions = service_info["workflow_type_versions"]
    available_wf_types: List[str] = list(map(str, wf_type_versions.keys()))
    if wf_type not in available_wf_types:
        abort(400, "The `workflow_type` is invalid.")
    available_wf_versions =\
        wf_type_versions[wf_type]["workflow_type_version"]  # type: ignore
    if wf_type_version not in available_wf_versions:
        abort(400, "The `workflow_type_version` is invalid.")


def validate_run_id(run_id: str) -> None:
    all_run_ids: List[str] = glob_all_run_ids()
    if run_id not in all_run_ids:
        abort(404, f"The run ID `{run_id}` does not exist.")


def validate_meta_characters(_type: str, content: str) -> None:
    """
    This function validates the string that will actually be evaluated in eval
    in run.sh. The possible types of strings are 'workflow_url',
    'workflow_engine_name', and 'workflow_engine_params'. If these strings
    contain any of the characters in the list of prohibited character types
    below, this will abort.

    In POST /runs, this is called as shown below:

    validate_meta_characters("workflow_engine_params", joined_params)
    validate_meta_characters("workflow_url", run_request["workflow_url"])
    validate_meta_characters("workflow_engine_name",
                             run_request["workflow_engine_name"])
    """
    prohibited_characters = [";", "!", "?", "(", ")", "[", "]", "{", "}", "*",
                             "\\", "&", r"`", "^", "<", ">", "|", "$"]
    for char in content:
        if char in prohibited_characters:
            abort(
                400, f"The `{_type}` contains a prohibited character `{char}`.")



#!/usr/bin/env python3
# coding: utf-8
from shutil import make_archive
from tempfile import NamedTemporaryFile
from typing import List
from uuid import uuid4

from flask import Blueprint, Response, abort, request, send_file
from flask.json import jsonify

from sapporo.config import str2bool
from sapporo.const import GET_STATUS_CODE, POST_STATUS_CODE
from sapporo.model import (ParseRequest, RunId, RunListResponse, RunLog,
                           RunRequest, RunStatus, ServiceInfo, Workflow)
from sapporo.model.factory import (generate_executable_workflows,
                                   generate_run_id, generate_run_list,
                                   generate_run_log, generate_run_status,
                                   generate_service_info)
from sapporo.parser import parse_workflows
from sapporo.run import (cancel_run, fork_run, path_hierarchy, prepare_run_dir,
                         resolve_requested_file_path)
from sapporo.validator import (validate_get_runs,
                               validate_post_parse_workflows, validate_run_id,
                               validate_run_request)

app_bp = Blueprint("sapporo", __name__)


@app_bp.route("/service-info", methods=["GET"])
def get_service_info() -> Response:
    res_body: ServiceInfo = generate_service_info()
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/executable-workflows", methods=["GET"])
def get_executable_workflows() -> Response:
    res_body: List[Workflow] = generate_executable_workflows()
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/parse-workflow", methods=["POST"])
def post_parse_workflows() -> Response:
    parse_request: ParseRequest = validate_post_parse_workflows()
    res_body = parse_workflows(parse_request)
    response: Response = jsonify(res_body)
    response.status_code = POST_STATUS_CODE

    return response


@app_bp.route("/runs", methods=["GET"])
def get_runs() -> Response:
    validate_get_runs()
    res_body: RunListResponse = generate_run_list()
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs", methods=["POST"])
def post_runs() -> Response:
    run_id = str(uuid4())
    run_request: RunRequest = validate_run_request(run_id)
    prepare_run_dir(run_id, run_request)
    fork_run(run_id)
    response: Response = jsonify(generate_run_id(run_id))
    response.status_code = POST_STATUS_CODE

    return response


@app_bp.route("/runs/<string:run_id>", methods=["GET"])
def get_runs_id(run_id: str) -> Response:
    validate_run_id(run_id)
    res_body: RunLog = generate_run_log(run_id)
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs/<string:run_id>/cancel", methods=["POST"])
def post_runs_id_cancel(run_id: str) -> Response:
    validate_run_id(run_id)
    cancel_run(run_id)
    res_body: RunId = generate_run_id(run_id)
    response: Response = jsonify(res_body)
    response.status_code = POST_STATUS_CODE

    return response


@app_bp.route("/runs/<string:run_id>/status", methods=["GET"])
def get_runs_id_status(run_id: str) -> Response:
    validate_run_id(run_id)
    res_body: RunStatus = generate_run_status(run_id)
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs/<string:run_id>/data/", methods=["GET"])
@app_bp.route("/runs/<string:run_id>/data/<path:subpath>", methods=["GET"])
def get_runs_id_data(run_id: str, subpath: str = "") -> Response:
    validate_run_id(run_id)
    requested_path = resolve_requested_file_path(run_id, subpath)
    if not requested_path.exists():
        abort(404, f"`{subpath}` is not found.")
    if requested_path.is_file():
        return send_file(requested_path, as_attachment=True)  # type: ignore
    if str2bool(request.args.get("download", False)):
        with NamedTemporaryFile() as f:
            archive = make_archive(f.name, "zip",
                                   root_dir=requested_path.parent,
                                   base_dir=requested_path.name)
            return send_file(archive, as_attachment=True,  # type: ignore
                             download_name=f"{requested_path.name}.zip")
    else:
        response: Response = \
            jsonify(path_hierarchy(requested_path, requested_path))
        response.status_code = GET_STATUS_CODE
        return response



#!/usr/bin/env python3
# coding: utf-8
import hashlib
import json
import os
import platform
import shlex
import shutil
import stat
import subprocess
import urllib
import urllib.request
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, TypedDict, cast
from urllib.parse import urlsplit

import magic
import psutil
import yaml
from rocrate.model.computationalworkflow import ComputationalWorkflow
from rocrate.model.computerlanguage import ComputerLanguage
from rocrate.model.computerlanguage import get_lang as ro_crate_get_lang
from rocrate.model.contextentity import ContextEntity
from rocrate.model.data_entity import DataEntity
from rocrate.model.dataset import Dataset
from rocrate.model.file import File
from rocrate.model.metadata import WORKFLOW_PROFILE, Metadata
from rocrate.model.root_dataset import RootDataset
from rocrate.model.softwareapplication import SoftwareApplication
from rocrate.model.testdefinition import TestDefinition
from rocrate.model.testinstance import TestInstance
from rocrate.model.testservice import TestService
from rocrate.model.testsuite import TestSuite
from rocrate.rocrate import ROCrate
from rocrate.utils import get_norm_value

from sapporo.const import RUN_DIR_STRUCTURE, RUN_DIR_STRUCTURE_KEYS
from sapporo.model import AttachedFile, RunRequest, ServiceInfo

TERM_PATH = Path(__file__).parent.joinpath("ro-terms.csv")
TERM_URL_BASE = "https://raw.githubusercontent.com/sapporo-wes/sapporo-service/main/sapporo/ro-terms.csv"
SAPPORO_EXTRA_TERMS: Dict[str, str] = {}
with TERM_PATH.open(mode="r", encoding="utf-8") as f:
    for line in f:
        if line.startswith("term"):
            continue
        term = line.strip().split(",")[0]
        SAPPORO_EXTRA_TERMS[term] = f"{TERM_URL_BASE}#{term}"

# https://www.researchobject.org/ro-terms/
# https://www.researchobject.org/ro-terms/#why-are-terms-collected-in-a-csv
TESTING_EXTRA_TERMS: Dict[str, str] = {
    "TestSuite": "https://w3id.org/ro/terms/test#TestSuite",
    "TestInstance": "https://w3id.org/ro/terms/test#TestInstance",
    "TestService": "https://w3id.org/ro/terms/test#TestService",
    "TestDefinition": "https://w3id.org/ro/terms/test#TestDefinition",
    "instance": "https://w3id.org/ro/terms/test#instance",
    "runsOn": "https://w3id.org/ro/terms/test#runsOn",
    "resource": "https://w3id.org/ro/terms/test#resource",
    "definition": "https://w3id.org/ro/terms/test#definition",
    "engineVersion": "https://w3id.org/ro/terms/test#engineVersion",
}

# https://www.researchobject.org/ro-crate/1.1/metadata.html#additional-metadata-standards
# - https://bioschemas.org/ComputationalWorkflow
# - https://bioschemas.org/FormalParameter
# - https://bioschemas.org/ComputationalWorkflow#input
# - https://bioschemas.org/ComputationalWorkflow#output

# === type definitions ===


class SapporoConfig(TypedDict):
    sapporo_version: str
    get_runs: bool
    workflow_attachment: bool
    registered_only_mode: bool
    service_info: str
    executable_workflows: str
    run_sh: str
    url_prefix: str
    sapporo_endpoint: str


class YevisAuthor(TypedDict):
    github_account: str
    name: str
    affiliation: str
    orcid: Optional[str]


class YevisLanguage(TypedDict):
    type: str
    version: str


class YevisFile(TypedDict):
    url: str
    target: str
    type: Literal["primary", "secondary"]


class YevisTestFile(TypedDict):
    url: str
    target: str
    type: Literal["wf_params", "wf_engine_params", "other"]


class YevisTest(TypedDict):
    id: str
    files: List[YevisTestFile]


class YevisWorkflow(TypedDict):
    name: str
    readme: str
    language: YevisLanguage
    files: List[YevisFile]
    testing: List[YevisTest]


class YevisMetadata(TypedDict):
    id: str
    version: str
    license: str
    authors: List[YevisAuthor]
    workflow: YevisWorkflow


# === functions ===


def generate_ro_crate(inputted_run_dir: str) -> None:
    """\
    Called in run.sh
    """
    run_dir: Path = Path(inputted_run_dir).resolve(strict=True)
    if not run_dir.is_dir():
        raise NotADirectoryError(f"{run_dir} is not a directory.")
    run_request: RunRequest = read_file(run_dir, "run_request")
    sapporo_config: SapporoConfig = read_file(run_dir, "sapporo_config")
    service_info: ServiceInfo = read_file(run_dir, "service_info")
    yevis_metadata: Optional[YevisMetadata] = read_file(run_dir, "yevis_metadata")
    run_id = run_dir.name

    crate = ROCrate(init=False, gen_preview=False)
    add_root_data_entity(crate, yevis_metadata)
    add_dataset_dir(crate, run_dir)
    add_workflow(crate, run_dir, run_request, yevis_metadata)
    add_workflow_attachment(crate, run_dir, run_request, yevis_metadata)
    add_test(crate, run_dir, run_request, sapporo_config, service_info, yevis_metadata, run_id)

    crate.write(run_dir)


def read_file(run_dir: Path, file_type: RUN_DIR_STRUCTURE_KEYS, one_line: bool = False, raw: bool = False) -> Any:
    if "dir" in file_type:
        return None
    file_path = run_dir.joinpath(RUN_DIR_STRUCTURE[file_type])
    if file_path.exists() is False:
        return None
    if file_path.is_file() is False:
        return None
    with file_path.open(mode="r", encoding="utf-8") as f:
        if one_line:
            return f.readline().strip()
        if raw:
            return f.read()
        try:
            return yaml.load(f, Loader=yaml.SafeLoader)
        except Exception:
            return f.read()


def add_root_data_entity(crate: ROCrate, yevis_meta: Optional[YevisMetadata]) -> None:
    """\
    Modified from crate.__init__from_tree()

    https://www.researchobject.org/ro-crate/1.1/root-data-entity.html#direct-properties-of-the-root-data-entity
    """
    root_dataset_ins = RootDataset(crate, properties={
        "name": "Sapporo RO-Crate",
        "description": "RO-Crate of Sapporo's run execution results.",
    })
    metadata_ins = Metadata(crate)
    metadata_ins.extra_terms.update(SAPPORO_EXTRA_TERMS)

    if yevis_meta is not None:
        root_dataset_ins.append_to("license", generate_license(crate, yevis_meta), compact=True)
        authors = yevis_meta["authors"]
        for author in authors:
            root_dataset_ins.append_to("author", generate_person(crate, author), compact=True)

    crate.add(
        root_dataset_ins,
        metadata_ins,
    )


def generate_license(crate: ROCrate, yevis_meta: YevisMetadata) -> ContextEntity:
    """\
    Call GitHub REST API to get license information.
    """
    license = yevis_meta["license"]
    with urllib.request.urlopen(f"https://api.github.com/licenses/{license}") as f:
        if f.status == 200:
            license_info = json.load(f)
            name = license_info["name"]
            id_ = license_info["html_url"]
        else:
            name = license
            id_ = license
    license_ins = ContextEntity(crate, id_, properties={
        "@type": ["CreativeWork"],
        "name": name,
    })

    crate.add(license_ins)

    return license_ins


def generate_person(crate: ROCrate, author: YevisAuthor) -> ContextEntity:
    if "orcid" in author:
        id_ = f"https://orcid.org/{author['orcid']}"
    else:
        id_ = f"https://github.com/{author['github_account']}"
    person_ins = ContextEntity(crate, id_, properties={
        "@type": ["Person"],
        "name": author["name"],
        "affiliation": author["affiliation"],
    })

    crate.add(person_ins)

    return person_ins


def add_dataset_dir(crate: ROCrate, run_dir: Path) -> None:
    exe_dir = run_dir.joinpath(RUN_DIR_STRUCTURE["exe_dir"])
    crate.add_dataset(
        exe_dir, exe_dir.relative_to(run_dir), {
            "name": "Sapporo execution directory",
        }
    )
    outputs_dir = run_dir.joinpath(RUN_DIR_STRUCTURE["outputs_dir"])
    crate.add_dataset(
        outputs_dir, outputs_dir.relative_to(run_dir), {
            "name": "Sapporo outputs directory",
        }
    )


def append_exe_dir_dataset(crate: ROCrate, ins: DataEntity) -> None:
    for entity in crate.get_entities():
        if isinstance(entity, Dataset):
            if str(entity["@id"]) == f"{RUN_DIR_STRUCTURE['exe_dir']}/":
                entity.append_to("hasPart", ins, compact=True)


def append_outputs_dir_dataset(crate: ROCrate, ins: DataEntity) -> None:
    for entity in crate.get_entities():
        if isinstance(entity, Dataset):
            if str(entity["@id"]) == f"{RUN_DIR_STRUCTURE['outputs_dir']}/":
                entity.append_to("hasPart", ins, compact=True)


def add_workflow(crate: ROCrate, run_dir: Path, run_request: RunRequest, yevis_meta: Optional[YevisMetadata]) -> None:
    """\
    Modified from crate.add_workflow()

    RunRequest:
      - wf_url: Remote location, or local file path attached as workflow_attachment and downloaded to exe_dir
    """
    wf_url = cast(str, run_request["workflow_url"])
    wf_url_parts = urlsplit(wf_url)
    if wf_url_parts.scheme == "http" or wf_url_parts.scheme == "https":
        tmp_file_path, _ = urllib.request.urlretrieve(wf_url)
        wf_file_path = Path(tmp_file_path)
        wf_ins = ComputationalWorkflow(crate, wf_url)
        update_local_file_stat(crate, wf_ins, wf_file_path)
        del wf_ins["dateModified"]
        del wf_ins["uid"]
        del wf_ins["gid"]
        del wf_ins["mode"]
    else:
        wf_file_path = run_dir.joinpath(RUN_DIR_STRUCTURE["exe_dir"], wf_url).resolve(strict=True)
        wf_ins = ComputationalWorkflow(crate, wf_file_path, wf_file_path.relative_to(run_dir))
        update_local_file_stat(crate, wf_ins, wf_file_path)
        append_exe_dir_dataset(crate, wf_ins)

    crate.add(wf_ins)

    if run_request["workflow_name"] is not None:
        wf_ins["name"] = run_request["workflow_name"]

    wf_ins.lang = generate_wf_lang(crate, run_request)

    crate.mainEntity = wf_ins
    profiles = set(_.rstrip("/") for _ in get_norm_value(crate.metadata, "conformsTo"))
    profiles.add(WORKFLOW_PROFILE)
    crate.metadata["conformsTo"] = [{"@id": _} for _ in sorted(profiles)]

    if yevis_meta is not None:
        wf_ins["yevisId"] = yevis_meta["id"]
        wf_ins["version"] = yevis_meta["version"]
        wf_ins["name"] = yevis_meta["workflow"]["name"]
        description_ins = ContextEntity(crate, yevis_meta["workflow"]["readme"], properties={
            "@type": ["WebPage"],
            "name": "README.md",
        })
        crate.add(description_ins)
        wf_ins["description"] = description_ins


def update_local_file_stat(crate: ROCrate, file_ins: File, file_path: Path, include_content: bool = True) -> None:
    if file_path.is_file() is False:
        return
    if file_path.exists() is False:
        return

    # From file stat
    stat_result = file_path.stat()

    # https://schema.org/MediaObject
    file_ins["contentSize"] = stat_result.st_size
    file_ins["dateModified"] = datetime.fromtimestamp(stat_result.st_mtime).isoformat()

    # additional properties (not defined)
    file_ins["uid"] = stat_result.st_uid
    file_ins["gid"] = stat_result.st_gid
    file_ins["mode"] = stat.filemode(stat_result.st_mode)

    # add file line count
    try:
        file_ins["lineCount"] = count_lines(file_path)
    except UnicodeDecodeError:
        pass

    # checksum using sha512 (https://www.researchobject.org/ro-crate/1.1/appendix/implementation-notes.html#combining-with-other-packaging-schemes)
    file_ins["sha512"] = generate_sha512(file_path)

    # https://pypi.org/project/python-magic/
    file_ins["encodingFormat"] = magic.from_file(file_path, mime=True)

    if include_content:
        # under 10kb, attach as text
        if file_ins["contentSize"] < 10 * 1024:
            try:
                file_ins["text"] = file_path.read_text()
            except UnicodeDecodeError:
                pass

    edam = inspect_edam_format(file_path)
    if edam is not None:
        edam_ins = ContextEntity(crate, edam["url"], properties={
            "@type": ["Format"],
            "name": edam["name"],
        })
        crate.add(edam_ins)
        file_ins.append_to("format", edam_ins, compact=True)


def count_lines(file_path: Path) -> int:
    block_size = 65536
    count = 0
    with file_path.open("r") as f:
        while True:
            buffer = f.read(block_size)
            if not buffer:
                break
            count += buffer.count("\n")

    return count


def generate_sha512(file_path: Path) -> str:
    block_size = 65536
    sha512 = hashlib.sha512()
    with file_path.open("rb") as f:
        while True:
            buffer = f.read(block_size)
            if not buffer:
                break
            sha512.update(buffer)

    return sha512.hexdigest()


class EDAM(TypedDict):
    url: str
    name: str


EDAM_MAPPING: Dict[str, EDAM] = {
    ".bam": {
        "url": "http://edamontology.org/format_2572",
        "name": "BAM format, the binary, BGZF-formatted compressed version of SAM format for alignment of nucleotide sequences (e.g. sequencing reads) to (a) reference sequence(s). May contain base-call and alignment qualities and other data.",
    },
    ".bb": {
        "url": "http://edamontology.org/format_3004",
        "name": "bigBed format for large sequence annotation tracks, similar to textual BED format.",
    },
    ".bed": {
        "url": "http://edamontology.org/format_3003",
        "name": "Browser Extensible Data (BED) format of sequence annotation track, typically to be displayed in a genome browser.",
    },
    ".bw": {
        "url": "http://edamontology.org/format_3006",
        "name": "bigWig format for large sequence annotation tracks that consist of a value for each sequence position. Similar to textual WIG format.",
    },
    ".fa": {
        "url": "http://edamontology.org/format_1929",
        "name": "FASTA format including NCBI-style IDs.",
    },
    ".fasta": {
        "url": "http://edamontology.org/format_1929",
        "name": "FASTA format including NCBI-style IDs.",
    },
    ".fq": {
        "url": "http://edamontology.org/format_1930",
        "name": "FASTQ short read format ignoring quality scores.",
    },
    ".fq.gz": {
        "url": "http://edamontology.org/format_1930",
        "name": "FASTQ short read format ignoring quality scores.",
    },
    ".gtf": {
        "url": "http://edamontology.org/format_2306",
        "name": "Gene Transfer Format (GTF), a restricted version of GFF.",
    },
    ".gff": {
        "url": "http://edamontology.org/format_1975",
        "name": "Generic Feature Format version 3 (GFF3) of sequence features.",
    },
    ".sam": {
        "url": "http://edamontology.org/format_2573",
        "name": "Sequence Alignment/Map (SAM) format for alignment of nucleotide sequences (e.g. sequencing reads) to (a) reference sequence(s). May contain base-call and alignment qualities and other data.",
    },
    ".vcf": {
        "url": "http://edamontology.org/format_3016",
        "name": "Variant Call Format (VCF) for sequence variation (indels, polymorphisms, structural variation).",
    },
    ".vcf.gz": {
        "url": "http://edamontology.org/format_3016",
        "name": "Variant Call Format (VCF) for sequence variation (indels, polymorphisms, structural variation).",
    },
    ".wig": {
        "url": "http://edamontology.org/format_3005",
        "name": "Wiggle format (WIG) of a sequence annotation track that consists of a value for each sequence position. Typically to be displayed in a genome browser.",
    },
}


def inspect_edam_format(file_path: Path) -> Optional[EDAM]:
    """\
    TODO: use tataki (https://github.com/suecharo/tataki)
    """
    for ext, edam in EDAM_MAPPING.items():
        if file_path.name.endswith(ext):
            return edam

    return None


def generate_wf_lang(crate: ROCrate, run_request: RunRequest) -> ComputerLanguage:
    """\
    wf_type: "CWL", "WDL", "NFL", "SMK" or others
    wf_type_version: str
    """
    wf_type = cast(str, run_request["workflow_type"])
    wf_type_version = cast(str, run_request["workflow_type_version"])

    lang_type_for_ro_crate = wf_type
    if wf_type.lower() == "nfl":
        lang_type_for_ro_crate = "nextflow"
    elif wf_type.lower() == "smk":
        lang_type_for_ro_crate = "snakemake"
    try:
        lang_ins = ro_crate_get_lang(crate, lang_type_for_ro_crate, wf_type_version)
        for filed in ["identifier", "url"]:
            id_ = get_norm_value(lang_ins, filed)[0]
            cxt = ContextEntity(crate, id_, properties={
                "@type": ["WebPage"],
            })
            crate.add(cxt)
            # lang_ins.append_to(filed, cxt) # bug of ro_crate_py
    except ValueError as e:
        if "Unknown language" in str(e):
            # case: WDL or others
            if wf_type.lower() == "wdl":
                id_ = "https://openwdl.org"
                lang_ins = ComputerLanguage(
                    crate,
                    id_,
                    properties={
                        "name": "Workflow Description Language",
                        "alternateName": "WDL",
                        "version": wf_type_version,
                    })
                ctx = ContextEntity(crate, id_, properties={
                    "@type": ["WebPage"],
                })
                lang_ins.append_to("identifier", ctx, compact=True)
                lang_ins.append_to("url", ctx, compact=True)
                crate.add(ctx)
            else:
                lang_ins = ComputerLanguage(
                    crate,
                    wf_type,
                    properties={
                        "name": wf_type,
                        "version": wf_type_version,
                    })
        else:
            raise e

    crate.add(lang_ins)

    return lang_ins


def add_workflow_attachment(crate: ROCrate, run_dir: Path, run_request: RunRequest,
                            yevis_meta: Optional[YevisMetadata]) -> None:
    """\
    If no Yevis (Sapporo only): All workflow attachments are treated as workflow inputs.
    If with Yevis: Workflow attachments are treated as workflow inputs, but test files are added to TestDefinition.

    workflow_attachment are placed in exe_dir (downloaded)
    """
    main_wf_id = crate.mainEntity["@id"]

    if yevis_meta is None:
        secondary_files: List[str] = []
    else:
        secondary_files = [file["target"] for file in yevis_meta["workflow"]["files"] if file["type"] == "secondary"]

    wf_attachment = cast(str, run_request["workflow_attachment"])  # encoded json string
    wf_attachment_obj: List[AttachedFile] = json.loads(wf_attachment)
    for item in wf_attachment_obj:
        if yevis_meta is not None:
            if item["file_name"] not in secondary_files:
                continue

        source = run_dir.joinpath(RUN_DIR_STRUCTURE["exe_dir"], item["file_name"])
        dest = source.relative_to(run_dir)
        if str(dest) == str(main_wf_id):
            continue
        type_list = ["File", "FormalParameter", "WorkflowAttachment"]
        if "script" in magic.from_file(source):
            type_list.append("SoftwareSourceCode")

        if yevis_meta is None:
            url = item["file_url"]
        else:
            file = [f for f in yevis_meta["workflow"]["files"] if f["target"] == item["file_name"]][0]
            url = file["url"]

        file_ins = File(crate, source, dest, properties={
            "@type": type_list,
            "url": url,
        })
        update_local_file_stat(crate, file_ins, source, include_content=False)
        append_exe_dir_dataset(crate, file_ins)
        crate.mainEntity.append_to("attachment", file_ins, compact=True)
        crate.add(file_ins)


def add_test(crate: ROCrate, run_dir: Path, run_request: RunRequest,
             sapporo_config: SapporoConfig, service_info: ServiceInfo, yevis_meta: Optional[YevisMetadata], run_id: str) -> None:
    suite_ins = generate_test_suite(crate)
    crate.root_dataset.append_to("about", suite_ins, compact=True)

    test_ins = generate_test_instance(crate, sapporo_config)
    suite_ins.append_to("instance", test_ins, compact=True)

    test_service_ins = generate_sapporo_service(crate, run_dir, sapporo_config)
    test_ins.service = test_service_ins  # runsOn

    test_env_ins = generate_test_env(crate)
    test_ins.append_to("environment", test_env_ins, compact=True)

    test_definition_ins = generate_test_definition(crate, run_dir, run_request, service_info, yevis_meta)
    suite_ins.definition = test_definition_ins

    test_result_ins = generate_test_result(crate, run_dir, run_id)
    suite_ins.append_to("result", test_result_ins, compact=True)

    crate.metadata.extra_terms.update(TESTING_EXTRA_TERMS)


def generate_test_suite(crate: ROCrate) -> TestInstance:
    """\
    Modified from crate.add_test_suite()

    TestSuite: A set of tests for a computational workflow
      - instance: Instances of a test suite
      - definition: Metadata describing how to run the test
    """
    suite_ins = TestSuite(crate, identifier="sapporo-test-suite")
    suite_ins.name = "Sapporo test suite"
    suite_ins["mainEntity"] = crate.mainEntity

    crate.add(suite_ins)

    return suite_ins


def generate_test_instance(crate: ROCrate, sapporo_config: SapporoConfig) -> TestInstance:
    """\
    Modified from crate.add_test_instance()

    TestInstance: A specific project to execute a test suite on a test service
      - runsOn: Service where the test instance is executed
    """
    test_ins = TestInstance(crate, identifier="sapporo-test-instance")
    test_ins.url = sapporo_config["sapporo_endpoint"]
    test_ins.name = "Sapporo test instance"

    crate.add(test_ins)

    return test_ins


def generate_sapporo_service(crate: ROCrate, run_dir: Path, sapporo_config: SapporoConfig) -> TestService:
    """\
    TestService: A software service where tests can be run
      - resource: Relative URL of the test project on the service
    """
    test_service_ins = TestService(crate, identifier="sapporo-service", properties={
        "@type": ["TestService", "SapporoService", "SoftwareApplication"],
        "name": "Sapporo-service",
        "version": sapporo_config["sapporo_version"],
        "resource": sapporo_config["url_prefix"]  # relative URL
    })
    crate.add(test_service_ins)
    sapporo_url_ins = ContextEntity(
        crate,
        "https://github.com/sapporo-wes/sapporo-service",
        properties={
            "@type": ["WebPage"],
        })
    test_service_ins.append_to("url", sapporo_url_ins, compact=True)
    crate.add(sapporo_url_ins)

    # Generate runtime parameters for Sapporo-service
    sapporo_conf_ins = ContextEntity(crate, identifier="sapporo-config", properties={
        "@type": ["SapporoConfiguration"],
        "getRuns": sapporo_config["get_runs"],
        "workflowAttachment": sapporo_config["workflow_attachment"],
        "registeredOnlyMode": sapporo_config["registered_only_mode"],
        "urlPrefix": sapporo_config["url_prefix"],
        "sapporoEndpoint": sapporo_config["sapporo_endpoint"],
    })
    test_service_ins.append_to("configuration", sapporo_conf_ins, compact=True)
    crate.add(sapporo_conf_ins)

    # Add local files about service-info, executable-workflows, run.sh, sapporo_config.json
    files: List[Tuple[RUN_DIR_STRUCTURE_KEYS, str, str]] = [
        ("service_info", "serviceInfo", "Sapporo service info"),
        ("executable_workflows", "executableWorkflows", "Sapporo executable workflows"),
        ("run_sh", "runSh", "Sapporo run.sh"),
        ("sapporo_config", "sapporoConfig", "Sapporo runtime configuration"),
    ]
    for key, field_key, name in files:
        source = run_dir.joinpath(RUN_DIR_STRUCTURE[key])
        if source.exists() is False:
            continue
        dest = source.relative_to(run_dir)
        file_ins = File(crate, source, dest, properties={
            "name": name,
        })
        update_local_file_stat(crate, file_ins, source)
        sapporo_conf_ins.append_to(field_key, file_ins, compact=True)
        crate.add(file_ins)

    return test_service_ins


def generate_test_env(crate: ROCrate) -> ContextEntity:
    """\
    Generate computational environment of the test instance
    """
    uname = platform.uname()
    in_docker = os.path.exists("/.dockerenv")
    test_env_ins = ContextEntity(crate, identifier="sapporo-test-environment", properties={
        "@type": ["TestEnvironment"],
        "os": uname.system,
        "osVersion": uname.release,
        "cpuArchitecture": uname.machine,
        "cpuCount": psutil.cpu_count(),
        "totalMemory": psutil.virtual_memory().total,
        "freeDiskSpace": psutil.disk_usage("/").free,
        "uid": os.getuid(),
        "gid": os.getgid(),
        "inDocker": in_docker,
    })

    crate.add(test_env_ins)

    return test_env_ins


def generate_test_definition(crate: ROCrate, run_dir: Path, run_request: RunRequest,
                             service_info: ServiceInfo, yevis_meta: Optional[YevisMetadata]) -> TestDefinition:
    """\
    Modified from crate.add_test_definition()
    """
    source = run_dir.joinpath(RUN_DIR_STRUCTURE["run_request"])
    dest = source.relative_to(run_dir)
    test_def_ins = TestDefinition(crate, source, dest, properties={
        "@type": ["TestDefinition", "File"],
        "name": "Sapporo run request",
    })
    update_local_file_stat(crate, test_def_ins, source)
    crate.add(test_def_ins)

    # workflow parameter files
    files: List[Tuple[RUN_DIR_STRUCTURE_KEYS, str, str]] = [
        ("wf_params", "workflowParameters", "Sapporo workflow parameters"),
        ("wf_engine_params", "workflowEngineParameters", "Sapporo workflow engine parameters"),
        ("yevis_metadata", "yevisMetadata", "Yevis metadata"),
    ]
    for key, field_key, name in files:
        source = run_dir.joinpath(RUN_DIR_STRUCTURE[key])
        if not source.exists():
            continue
        dest = source.relative_to(run_dir)
        file_ins = File(crate, source, dest, properties={
            "@type": ["File"],
            "name": name,
        })
        if yevis_meta is not None:
            file_obj = [f for f in yevis_meta["workflow"]["testing"][0]["files"] if f["type"] == key]
            if len(file_obj) > 0:
                file_ins["url"] = file_obj[0]["url"]
        update_local_file_stat(crate, file_ins, source)
        if key == "wf_params":
            append_exe_dir_dataset(crate, file_ins)
        test_def_ins.append_to(field_key, file_ins, compact=True)
        crate.add(file_ins)

    # cmd
    cmd_str = read_file(run_dir, "cmd", one_line=True)
    test_def_ins["cmd"] = cmd_str

    # workflow engine
    wf_engine_name = run_request["workflow_engine_name"]
    wf_engine_version = service_info["workflow_engine_versions"][wf_engine_name]  # validated at run acceptance
    wf_engine_ins = generate_wf_engine(crate, wf_engine_name, wf_engine_version)
    test_def_ins.engine = wf_engine_ins
    test_def_ins.engineVersion = wf_engine_version

    if yevis_meta is not None:
        test_def_ins["yevisTestId"] = yevis_meta["workflow"]["testing"][0]["id"]

        # add test data
        test_files = [f for f in yevis_meta["workflow"]["testing"][0]["files"] if f["type"] == "other"]
        for test_file in test_files:
            source = run_dir.joinpath(RUN_DIR_STRUCTURE["exe_dir"], test_file["target"])
            if source.exists() is False:
                continue
            dest = source.relative_to(run_dir)
            file_ins = File(crate, source, dest, properties={
                "@type": ["File", "FormalParameter"],
                "url": test_file["url"],
            })
            update_local_file_stat(crate, file_ins, source)
            append_exe_dir_dataset(crate, file_ins)
            test_def_ins.append_to("input", file_ins, compact=True)
            crate.add(file_ins)

    return test_def_ins


def generate_wf_engine(crate: ROCrate, wf_engine_name: str, wf_engine_version: str) -> SoftwareApplication:
    urls = {
        "cwltool": "https://github.com/common-workflow-language/cwltool",
        "cromwell": "https://cromwell.readthedocs.io/en/stable/",
        "nextflow": "https://www.nextflow.io",
        "snakemake": "https://snakemake.readthedocs.io/en/stable/",
    }
    engine = SoftwareApplication(
        crate,
        identifier=wf_engine_name,
        properties={
            "name": wf_engine_name,
            "version": wf_engine_version,
        })
    if urls.get(wf_engine_name) is not None:
        ctx = ContextEntity(crate, urls[wf_engine_name], properties={
            "@type": ["WebPage"],
        })
        engine.append_to("url", ctx, compact=True)
        crate.add(ctx)

    crate.add(engine)

    return engine


def generate_test_result(crate: ROCrate, run_dir: Path, run_id: str) -> ContextEntity:
    test_result_ins = ContextEntity(crate, identifier="sapporo-test-result", properties={
        "@type": ["TestResult"],
        "name": "Sapporo test result",
    })
    crate.add(test_result_ins)

    test_result_ins["runId"] = run_id

    # Add one-line text files
    one_line_files: List[Tuple[RUN_DIR_STRUCTURE_KEYS, str]] = [
        ("start_time", "startTime"),
        ("end_time", "endTime"),
        ("exit_code", "exitCode"),
        ("pid", "pid"),
        ("state", "state"),
    ]
    for key, field_key in one_line_files:
        content = read_file(run_dir, key, one_line=True)
        if content is None:
            continue
        if key == "pid" or key == "exit_code":
            test_result_ins[field_key] = int(content)
        else:
            test_result_ins[field_key] = content

    # Add log files
    log_files: List[Tuple[RUN_DIR_STRUCTURE_KEYS, str, str]] = [
        ("stdout", "stdout", "Sapporo stdout"),
        ("stderr", "stderr", "Sapporo stderr"),
        ("task_logs", "taskLogs", "Sapporo task logs"),
    ]
    for key, field_key, name in log_files:
        source = run_dir.joinpath(RUN_DIR_STRUCTURE[key])
        if source.exists() is False:
            continue
        dest = source.relative_to(run_dir)
        file_ins = File(crate, source, dest, properties={
            "name": name,
        })
        update_local_file_stat(crate, file_ins, source)
        test_result_ins.append_to(field_key, file_ins, compact=True)
        crate.add(file_ins)

    # Add output files
    outputs: Optional[List[AttachedFile]] = read_file(run_dir, "outputs")
    for source in run_dir.joinpath(RUN_DIR_STRUCTURE["outputs_dir"]).glob("**/*"):
        if source.is_dir():
            continue
        source = source.resolve(strict=True)
        dest = source.relative_to(run_dir)
        file_ins = File(crate, source, dest, properties={
            "@type": ["File", "FormalParameter", "OutputFile"],
        })
        update_local_file_stat(crate, file_ins, source)

        if outputs is not None:
            # Include the URL of Sapporo's download feature
            output_dir_dest = source.relative_to(run_dir.joinpath(RUN_DIR_STRUCTURE["outputs_dir"]))
            for output in outputs:
                if str(output["file_name"]) == str(output_dir_dest):
                    file_ins["url"] = output["file_url"]

        add_file_stats(crate, file_ins)

        append_outputs_dir_dataset(crate, file_ins)
        test_result_ins.append_to("outputs", file_ins, compact=True)
        crate.add(file_ins)

    # Add intermediate files
    test_result_ins["intermediateFiles"] = []
    already_added_ids = extract_exe_dir_file_ids(crate)
    for source in run_dir.joinpath(RUN_DIR_STRUCTURE["exe_dir"]).glob("**/*"):
        if source.is_dir():
            continue
        source = source.resolve(strict=True)
        dest = source.relative_to(run_dir)
        if str(dest) in already_added_ids:
            continue
        file_ins = File(crate, source, dest, properties={
            "@type": ["File", "FormalParameter", "IntermediateFile"],
        })
        update_local_file_stat(crate, file_ins, source, include_content=False)
        append_exe_dir_dataset(crate, file_ins)
        test_result_ins.append_to("intermediateFiles", file_ins, compact=True)
        crate.add(file_ins)

    return test_result_ins


def extract_exe_dir_file_ids(crate: ROCrate) -> List[str]:
    for entity in crate.get_entities():
        if isinstance(entity, Dataset):
            if str(entity["@id"]) == f"{RUN_DIR_STRUCTURE['exe_dir']}/":
                return cast(List[str], get_norm_value(entity, "hasPart"))
    return []


def add_file_stats(crate: ROCrate, file_ins: File) -> None:
    """\
    see "format" field of file_ins

    ".bam": "http://edamontology.org/format_2572"
    ".sam": "http://edamontology.org/format_2573",
      -> quay.io/biocontainers/samtools:1.15.1--h1170115_0
    ".vcf": "http://edamontology.org/format_3016",
      -> quay.io/biocontainers/vcftools:0.1.16--pl5321h9a82719_6
    """
    # TODO: use docker or local command?
    if shutil.which("docker") is None:
        return

    formats = get_norm_value(file_ins, "format")
    for format_ in formats:
        if format_ == "http://edamontology.org/format_2572" or format_ == "http://edamontology.org/format_2573":
            # bam or sam
            add_samtools_stats(crate, file_ins)
        elif format_ == "http://edamontology.org/format_3016":
            # vcf
            add_vcftools_stats(crate, file_ins)


def add_samtools_stats(crate: ROCrate, file_ins: File) -> None:
    """\
    $ samtools flagstats --output-fmt json <file_path>

    Using: quay.io/biocontainers/samtools:1.15.1--h1170115_0
    """
    source = file_ins.source
    cmd = shlex.split(" ".join([
        "docker",
        "run",
        "--rm",
        "-v",
        f"{source}:/work/{source.name}",
        "-w",
        "/work",
        "quay.io/biocontainers/samtools:1.15.1--h1170115_0",
        "samtools",
        "flagstats",
        "--output-fmt",
        "json",
        source.name,
    ]))
    proc = subprocess.run(cmd, capture_output=True)
    if proc.returncode != 0:
        return
    try:
        stats = json.loads(proc.stdout)
        total = stats["QC-passed reads"]["total"]
        mapped = stats["QC-passed reads"]["mapped"]
        unmapped = total - mapped
        duplicate = stats["QC-passed reads"]["duplicates"]
        stats_ins = ContextEntity(crate, properties={
            "@type": ["FileStats"],
            "totalReads": total,
            "mappedReads": mapped,
            "unmappedReads": unmapped,
            "duplicateReads": duplicate,
            "mappedRate": mapped / total,
            "unmappedRate": unmapped / total,
            "duplicateRate": duplicate / total,
        })
        stats_ins.append_to("generatedBy", find_or_generate_software_ins(crate, "samtools", "1.15.1--h1170115_0"), compact=True)
        file_ins.append_to("stats", stats_ins, compact=True)
        crate.add(stats_ins)
    except json.JSONDecodeError:
        return


def add_vcftools_stats(crate: ROCrate, file_ins: File) -> None:
    """\
    $ vcf-stats <file_path>

    Using: quay.io/biocontainers/vcftools:0.1.16--pl5321h9a82719_6
    """
    source = file_ins.source
    cmd = shlex.split(" ".join([
        "docker",
        "run",
        "--rm",
        "-v",
        f"{source}:/work/{source.name}",
        "-w",
        "/work",
        "quay.io/biocontainers/vcftools:0.1.16--pl5321h9a82719_6",
        "vcf-stats",
        source.name,
    ]))
    proc = subprocess.run(cmd, capture_output=True)
    if proc.returncode != 0:
        return
    try:
        stdout = proc.stdout.decode()
        stdout = stdout.strip()
        stdout = stdout.lstrip("$VAR1 = ")
        stdout = stdout.rstrip(";")
        stdout = stdout.replace("=>", ":")
        stdout = stdout.replace("\'", "\"")
        stats = json.loads(stdout)
        stats_ins = ContextEntity(crate, properties={
            "@type": ["FileStats"],
            "variantCount": stats["all"].get("count", 0),
            "snpsCount": stats["all"].get("snp_count", 0),
            "indelsCount": stats["all"].get("indel_count", 0),
        })
        stats_ins.append_to("generatedBy", find_or_generate_software_ins(crate, "vcftools", "0.1.16--pl5321h9a82719_6"), compact=True)
        file_ins.append_to("stats", stats_ins, compact=True)
        crate.add(stats_ins)
    except json.JSONDecodeError:
        return


def find_or_generate_software_ins(crate: ROCrate, name: str, version: str) -> SoftwareApplication:
    for entity in crate.get_entities():
        if isinstance(entity, SoftwareApplication):
            if entity["name"] == name:
                return entity
    software_ins = SoftwareApplication(crate, identifier=name, properties={
        "name": name,
        "version": version
    })
    crate.add(software_ins)

    return software_ins


if __name__ == "__main__":
    import sys
    inputted_dir = Path(sys.argv[1]).resolve(strict=True)
    generate_ro_crate(str(inputted_dir))



#!/usr/bin/env python3
# coding: utf-8
from flask import Flask

from sapporo.app import create_app, get_config
from sapporo.config import Config

config: Config = get_config()
app: Flask = create_app(config)



#!/usr/bin/env python3
# coding: utf-8
import sys
from traceback import format_exc

from flask import Flask, Response, current_app, jsonify
from flask_cors import CORS
from werkzeug.exceptions import HTTPException

from sapporo.config import (Config, TypedNamespace, get_config, parse_args,
                            validate_config)
from sapporo.controller import app_bp
from sapporo.model import ErrorResponse


def fix_errorhandler(app: Flask) -> Flask:
    @app.errorhandler(HTTPException)
    def error_handler(error: HTTPException) -> Response:
        res_body: ErrorResponse = {
            "msg": error.description or "",
            "status_code": error.code or 500,
        }
        response: Response = jsonify(res_body)
        response.status_code = error.code or 500
        return response

    @app.errorhandler(Exception)
    def error_handler_exception(exception: Exception) -> Response:
        current_app.logger.error(exception.args[0])
        current_app.logger.debug(format_exc())
        res_body: ErrorResponse = {
            "msg": "The server encountered an internal error and was unable to complete your request.",
            "status_code": 500,
        }
        if current_app.config["TESTING"]:
            res_body["msg"] = format_exc()
        response: Response = jsonify(res_body)
        response.status_code = 500
        return response

    return app


def create_app(config: Config) -> Flask:
    validate_config(config)
    app: Flask = Flask(__name__)
    app.register_blueprint(app_bp, url_prefix=config["url_prefix"])
    fix_errorhandler(app)
    CORS(app, resources={
         r"/*": {"origins": config["access_control_allow_origin"]}})
    app.config["RUN_DIR"] = config["run_dir"]
    app.config["SAPPORO_VERSION"] = config["sapporo_version"]
    app.config["GET_RUNS"] = config["get_runs"]
    app.config["WORKFLOW_ATTACHMENT"] = config["workflow_attachment"]
    app.config["REGISTERED_ONLY_MODE"] = config["registered_only_mode"]
    app.config["SERVICE_INFO"] = config["service_info"]
    app.config["EXECUTABLE_WORKFLOWS"] = config["executable_workflows"]
    app.config["RUN_SH"] = config["run_sh"]
    app.config["URL_PREFIX"] = config["url_prefix"]
    if config["debug"]:
        app.config["FLASK_ENV"] = "development"
        app.config["DEBUG"] = True
        app.config["TESTING"] = True
        app.logger.debug(f"config: {config}")

    return app


def main() -> None:
    args: TypedNamespace = parse_args(sys.argv[1:])
    config: Config = get_config(args)
    app: Flask = create_app(config)
    app.run(
        host=config["host"],
        port=config["port"],
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
# flake8: noqa: F401
from sapporo.model.sapporo_wes_1_0_1 import (AttachedFile,
                                             DefaultWorkflowEngineParameter,
                                             ErrorResponse, Log, ParseRequest,
                                             ParseResult, RunId,
                                             RunListResponse, RunLog,
                                             RunRequest, RunStatus,
                                             SecondaryFile, ServiceInfo, State,
                                             Workflow, WorkflowInput,
                                             WorkflowTypes,
                                             WorkflowTypeVersion)



#!/usr/bin/env python3
# coding: utf-8
from typing import Any, List

import requests

from sapporo.model import Workflow


def get_wfs(trs_endpoint: str) -> List[Workflow]:
    try:
        if not is_trs_endpoint(trs_endpoint):
            raise ValueError(f"Invalid TRS endpoint: {trs_endpoint}")
    except Exception as e:
        raise ValueError(f"Invalid TRS endpoint: {trs_endpoint} ({e})")
    try:
        return get_wfs_via_api(trs_endpoint)
    except Exception as e:
        raise ValueError(f"Failed to get TRS workflow list ({e})")


def get_service_info(trs_endpoint: str) -> Any:
    url = trs_endpoint.rstrip("/") + "/service-info"
    res = requests.get(url, allow_redirects=True)
    if res.status_code != 200:
        raise Exception("Failed to get service info")
    return res.json()


def is_trs_endpoint(trs_endpoint: str) -> bool:
    service_info = get_service_info(trs_endpoint)
    try:
        _type = service_info["type"]
        artifact = _type["artifact"]
        version = _type["version"]
    except Exception:
        raise ValueError(f"Invalid TRS endpoint: {trs_endpoint}")
    if artifact.lower() in ["trs", "gh-trs", "yevis"] and version.startswith("2."):
        return True
    return False


def get_tools(trs_endpoint: str) -> Any:
    url = trs_endpoint.rstrip("/") + "/tools"
    res = requests.get(url, allow_redirects=True)
    if res.status_code != 200:
        raise Exception("Failed to get tools")
    return res.json()


def get_wfs_via_api(trs_endpoint: str) -> List[Workflow]:
    tools = get_tools(trs_endpoint)
    wfs = []
    for tool in tools:
        try:
            _id = tool["id"]
            for ver in tool["versions"]:
                name = ver["name"]
                version = ver["id"]
                for _type in ver["descriptor_type"]:
                    wf: Workflow = {
                        "workflow_name": name + f": {version}",
                        "workflow_url": "",
                        "workflow_type": fix_wf_type(_type),
                        "workflow_type_version": ret_wf_type_version(_type),
                        "workflow_attachment": []
                    }
                    files = get_files_via_api(trs_endpoint, _id, version, _type)
                    for file in files:
                        url = file["path"]
                        if file["file_type"] == "PRIMARY_DESCRIPTOR":
                            wf["workflow_url"] = url
                        elif file["file_type"] == "SECONDARY_DESCRIPTOR":
                            wf["workflow_attachment"].append({
                                "file_name": url.split("/")[-1],
                                "file_url": url
                            })
                    if wf["workflow_url"] != "":
                        wfs.append(wf)
        except Exception:
            pass
    return wfs


def get_files_via_api(trs_endpoint: str, id: str, ver: str, _type: str) -> Any:
    url = trs_endpoint.rstrip("/") + f"/tools/{id}/versions/{ver}/{_type}/files"
    res = requests.get(url, allow_redirects=True)
    if res.status_code != 200:
        raise Exception("Failed to get files")
    return res.json()


def fix_wf_type(_type: str) -> str:
    _type = _type.lower()
    if _type == "cwl":
        return "CWL"
    elif _type == "wdl":
        return "WDL"
    elif _type == "nfl" or _type == "nextflow":
        return "NFL"
    elif _type == "smk" or _type == "snakemake":
        return "SMK"
    else:
        return _type


def ret_wf_type_version(_type: str) -> str:
    _type = _type.lower()
    if _type == "cwl":
        return "v1.0"
    elif _type == "wdl":
        return "1.0"
    elif _type == "nfl" or _type == "nextflow":
        return "1.0"
    elif _type == "smk" or _type == "snakemake":
        return "1.0"
    else:
        return "1.0"



#!/usr/bin/env python3
# coding: utf-8
# pylint: disable=no-else-return
import re
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from cwl_inputs_parser.utils import Inputs, as_uri
from cwl_inputs_parser.utils import \
    cwl_make_template as inputs_parser_make_template
from cwl_inputs_parser.utils import (download_file, is_remote_url,
                                     wf_location_to_inputs)
from cwl_utils.parser import cwl_version, load_document_by_string
from flask import abort
from schema_salad.utils import yaml_no_ts

from sapporo.model import ParseRequest, ParseResult


def parse_workflows(parse_request: ParseRequest) -> ParseResult:
    if parse_request["workflow_location"] is not None:
        if not is_remote_url(parse_request["workflow_location"]):
            abort(400, "Workflow location must be a remote URL")
        wf_content = download_file(parse_request["workflow_location"])
    else:
        wf_content = parse_request["workflow_content"]
    wf_location = parse_request["workflow_location"] or "."
    types_of_parsing = parse_request["types_of_parsing"] or [
        "workflow_type", "workflow_type_version"]

    wf_type = inspect_wf_type(wf_content, wf_location)
    wf_version = inspect_wf_version(wf_content, wf_type)

    inputs = None
    if wf_type == "CWL":
        if "make_template" in types_of_parsing:
            inputs = cwl_make_template(wf_content, wf_location)
        else:
            if "inputs" in types_of_parsing:
                try:
                    inputs = parse_cwl_inputs(
                        wf_content, wf_location)  # type: ignore
                except Exception:
                    inputs = cwl_make_template(wf_content, wf_location)
    else:
        if "inputs" in types_of_parsing or "make_template" in types_of_parsing:
            abort(
                400, f"Workflow type: `{wf_type}` is not supported parsing inputs or make template")

    parse_result: ParseResult = {
        "workflow_type": wf_type,  # type: ignore
        "workflow_type_version": wf_version,
        "inputs": inputs,
    }

    return parse_result


WF_TYPES = Literal["CWL", "WDL", "NFL", "SMK", "StreamFlow", "unknown"]


def inspect_wf_type(wf_content: str, wf_location: str) -> WF_TYPES:
    wf_type = check_by_shebang(wf_content)
    if wf_type != "unknown":
        return wf_type

    wf_type = check_by_cwl_utils(wf_content, wf_location)
    if wf_type != "unknown":
        return wf_type

    wf_type = check_by_regexp(wf_content)
    if wf_type != "unknown":
        return wf_type

    return "unknown"


def check_by_shebang(wf_content: str) -> WF_TYPES:
    first_line = wf_content.split("\n")[0]
    if first_line.startswith("#!"):
        if "cwl" in first_line:
            return "CWL"
        elif "nextflow" in first_line:
            return "NFL"
        elif "snakemake" in first_line:
            return "SMK"
        elif "cromwell" in first_line:
            return "WDL"
        elif "streamflow" in first_line:
            return "StreamFlow"

    return "unknown"


def check_by_cwl_utils(wf_content: str, wf_location: str) -> Literal["CWL", "unknown"]:
    try:
        load_document_by_string(wf_content, as_uri(wf_location))
        return "CWL"
    except Exception:
        return "unknown"


PATTERN_WDL = re.compile(r"^(workflow|task) \w* \{$")
PATTERN_SMK = re.compile(r"^rule \w*:$")
PATTERN_NFL = re.compile(r"^process \w* \{$")


def check_by_regexp(wf_content: str) -> WF_TYPES:
    for line in wf_content.split("\n"):
        if PATTERN_WDL.match(line):
            return "WDL"
        elif PATTERN_SMK.match(line):
            return "SMK"
        elif PATTERN_NFL.match(line):
            return "NFL"

    return "unknown"


def inspect_wf_version(wf_content: str, wf_type: WF_TYPES) -> str:
    wf_version = "unknown"
    if wf_type == "CWL":
        wf_version = inspect_cwl_version(wf_content)
    elif wf_type == "WDL":
        wf_version = inspect_wdl_version(wf_content)
    elif wf_type == "NFL":
        wf_version = inspect_nfl_version(wf_content)
    elif wf_type == "SMK":
        wf_version = inspect_smk_version()
    elif wf_type == "StreamFlow":
        wf_version = inspect_streamflow_version(wf_content)

    return wf_version


def inspect_cwl_version(wf_content: str) -> str:
    """
    https://www.commonwl.org/v1.2/CommandLineTool.html#CWLVersion
    """
    default_cwl_version = "v1.0"

    yaml = yaml_no_ts()
    yaml_obj = yaml.load(wf_content)

    return cwl_version(yaml_obj) or default_cwl_version


PATTERN_WDL_VERSION = re.compile(r"^version \d\.\d$")


def inspect_wdl_version(wf_content: str) -> str:
    default_wdl_version = "1.0"

    for line in wf_content.split("\n"):
        if PATTERN_WDL_VERSION.match(line):
            return line.split(" ")[1]

    return default_wdl_version


def inspect_nfl_version(wf_content: str) -> str:
    default_nfl_version = "1.0"

    for line in wf_content.split("\n"):
        if line == "nextflow.enable.dsl=2":
            return "DSL2"

    return default_nfl_version


def inspect_smk_version() -> str:
    default_smk_version = "1.0"

    return default_smk_version


def inspect_streamflow_version(wf_content: str) -> str:
    default_streamflow_version = "v1.0"

    yaml = yaml_no_ts()
    yaml_obj = yaml.load(wf_content)

    return yaml_obj['version'] or default_streamflow_version


def parse_cwl_inputs(wf_content: str, wf_location: str) -> List[Dict[str, Any]]:
    if is_remote_url(wf_location):
        inputs = wf_location_to_inputs(wf_location)
    else:
        wf_obj = load_document_by_string(wf_content, uri=Path.cwd().as_uri())
        inputs = Inputs(wf_obj)

    return inputs.as_dict()  # type: ignore


def cwl_make_template(wf_content: str, wf_location: str) -> Optional[str]:
    inputs: Optional[str] = None
    if is_remote_url(wf_location):
        inputs = inputs_parser_make_template(wf_location)
    else:
        with tempfile.NamedTemporaryFile(suffix=".cwl") as temp_file:
            temp_file.write(wf_content.encode("utf-8"))
            temp_file.flush()
            inputs = inputs_parser_make_template(temp_file.name)

    return inputs



#!/usr/bin/env python3
# coding: utf-8
import argparse
import json
import os
from argparse import ArgumentParser, Namespace
from pathlib import Path
from sys import version_info
from typing import List, Optional, Tuple, Union, cast

import pkg_resources
from jsonschema import validate

from sapporo.const import (DEFAULT_ACCESS_CONTROL_ALLOW_ORIGIN,
                           DEFAULT_EXECUTABLE_WORKFLOWS, DEFAULT_HOST,
                           DEFAULT_PORT, DEFAULT_RUN_DIR, DEFAULT_RUN_SH,
                           DEFAULT_SERVICE_INFO, DEFAULT_URL_PREFIX,
                           EXECUTABLE_WORKFLOWS_SCHEMA, SERVICE_INFO_SCHEMA)
from sapporo.model import Workflow

if version_info.minor < 8:
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


def str2bool(val: Union[str, bool]) -> bool:
    if isinstance(val, bool):
        return val
    if val.lower() in ["true", "yes", "y"]:
        return True
    if val.lower() in ["false", "no", "n"]:
        return False

    return bool(val)


class TypedNamespace(Namespace):
    host: Optional[str]
    port: Optional[int]
    debug: bool
    run_dir: Optional[Path]
    disable_get_runs: bool
    disable_workflow_attachment: bool
    run_only_registered_workflows: bool
    service_info: Optional[Path]
    executable_workflows: Optional[Path]
    run_sh: Optional[Path]
    url_prefix: Optional[str]


def parse_args(args: Optional[List[str]] = None) -> TypedNamespace:
    parser: ArgumentParser = argparse.ArgumentParser(
        description="Implementation of a GA4GH workflow execution service that can easily support various workflow runners.")

    parser.add_argument(
        "--host",
        metavar="",
        help=f"Host address of Flask. (default: {DEFAULT_HOST})"
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        metavar="",
        help=f"Port of Flask. (default: {DEFAULT_PORT})"
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug mode of Flask."
    )
    parser.add_argument(
        "-r",
        "--run-dir",
        type=Path,
        metavar="",
        help="Specify the run dir. (default: ./run)"
    )
    parser.add_argument(
        "--disable-get-runs",
        action="store_true",
        help="Disable endpoint of `GET /runs`."
    )
    parser.add_argument(
        "--disable-workflow-attachment",
        action="store_true",
        help="Disable `workflow_attachment` on endpoint `Post /runs`."
    )
    parser.add_argument(
        "--run-only-registered-workflows",
        action="store_true",
        help="Run only registered workflows. Check the registered workflows using `GET /executable-workflows`, and specify `workflow_name` in the `POST /run`."
    )
    parser.add_argument(
        "--service-info",
        type=Path,
        metavar="",
        help="Specify `service-info.json`. The `supported_wes_versions` and `system_state_counts` are overwritten in the application."
    )
    parser.add_argument(
        "--executable-workflows",
        type=Path,
        metavar="",
        help="Specify `executable-workflows.json`."
    )
    parser.add_argument(
        "--run-sh",
        type=Path,
        metavar="",
        help="Specify `run.sh`."
    )
    parser.add_argument(
        "--url-prefix",
        metavar="",
        help="Specify the prefix of the url (e.g. --url-prefix /foo -> /foo/service-info)."
    )

    if args is None:
        return cast(TypedNamespace, parser.parse_args())

    return cast(TypedNamespace, parser.parse_args(args))


class Config(TypedDict):
    host: str
    port: int
    debug: bool
    run_dir: Path
    sapporo_version: str
    get_runs: bool
    workflow_attachment: bool
    registered_only_mode: bool
    service_info: Path
    executable_workflows: Path
    run_sh: Path
    url_prefix: str
    access_control_allow_origin: str


def resolve_path_from_cwd(path: Union[Path, Tuple[Path]]) -> Path:
    if isinstance(path, tuple):
        path = path[0]
    if path.is_absolute():
        return path
    return path.resolve(strict=True)


def get_config(args: Optional[TypedNamespace] = None) -> Config:
    if args is None:
        args = parse_args()

    run_dir = args.run_dir or Path(os.environ.get("SAPPORO_RUN_DIR", DEFAULT_RUN_DIR))
    service_info = args.service_info or Path(os.environ.get("SAPPORO_SERVICE_INFO", DEFAULT_SERVICE_INFO))
    executable_workflows = args.executable_workflows or Path(os.environ.get("SAPPORO_EXECUTABLE_WORKFLOWS", DEFAULT_EXECUTABLE_WORKFLOWS))
    run_sh = args.run_sh or Path(os.environ.get("SAPPORO_RUN_SH", DEFAULT_RUN_SH))

    return {
        "host": args.host or str(os.environ.get("SAPPORO_HOST", DEFAULT_HOST)),
        "port": args.port or int(os.environ.get("SAPPORO_PORT", DEFAULT_PORT)),
        "debug": args.debug or str2bool(os.environ.get("SAPPORO_DEBUG", False)),
        "run_dir": resolve_path_from_cwd(run_dir),
        "sapporo_version": pkg_resources.get_distribution("sapporo").version,
        "get_runs": False if args.disable_get_runs else str2bool(os.environ.get("SAPPORO_GET_RUNS", True)),
        "workflow_attachment": False if args.disable_workflow_attachment else str2bool(os.environ.get("SAPPORO_WORKFLOW_ATTACHMENT", True)),
        "registered_only_mode": args.run_only_registered_workflows or str2bool(os.environ.get("SAPPORO_RUN_ONLY_REGISTERED_WORKFLOWS", False)),
        "service_info": resolve_path_from_cwd(service_info),
        "executable_workflows": resolve_path_from_cwd(executable_workflows),
        "run_sh": resolve_path_from_cwd(run_sh),
        "url_prefix": args.url_prefix or str(os.environ.get("SAPPORO_URL_PREFIX", DEFAULT_URL_PREFIX)),
        "access_control_allow_origin": os.environ.get("SAPPORO_ACCESS_CONTROL_ALLOW_ORIGIN", DEFAULT_ACCESS_CONTROL_ALLOW_ORIGIN)
    }


def validate_config(config: Config) -> None:
    if not config["service_info"].exists():
        raise ValueError(f"{config['service_info']} does not exist.")
    with config["service_info"].open(mode="r", encoding="utf-8") as f_data, \
            SERVICE_INFO_SCHEMA.open(mode="r", encoding="utf-8") as f_schema:
        validate(json.load(f_data), json.load(f_schema))

    if not config["executable_workflows"].exists():
        raise ValueError(f"{config['executable_workflows']} does not exist.")
    with config["executable_workflows"].open(mode="r", encoding="utf-8") as f_data, \
            EXECUTABLE_WORKFLOWS_SCHEMA.open(mode="r", encoding="utf-8") as f_schema:
        validate(json.load(f_data), json.load(f_schema))

    # Check uniqueness of workflow_name
    with config["executable_workflows"].open(mode="r", encoding="utf-8") as f_data:
        executable_wfs: List[Workflow] = json.load(f_data)["workflow"]
    wf_names: List[str] = [wf["workflow_name"] for wf in executable_wfs]
    if len(wf_names) != len(set(wf_names)):
        raise Exception(
            "The workflow name included in `executable-workflows.json` must be unique.")

    if not config["run_sh"].exists():
        raise ValueError(f"{config['run_sh']} does not exist.")



#!/usr/bin/env python3
# coding: utf-8
import collections
import json
import os
import shlex
import shutil
import signal
from pathlib import Path, PurePath
from subprocess import Popen
from typing import Any, Dict, Iterable, List, Optional
from unicodedata import normalize
from urllib import parse

import requests
from flask import current_app, request
from werkzeug.utils import _filename_ascii_strip_re

from sapporo.const import RUN_DIR_STRUCTURE, RUN_DIR_STRUCTURE_KEYS
from sapporo.model import RunRequest, State
from sapporo.model.factory import (generate_executable_workflows,
                                   generate_service_info)
from sapporo.model.sapporo_wes_1_0_1 import AttachedFile


def resolve_run_dir_path(run_id: str) -> Path:
    run_base_dir: Path = current_app.config["RUN_DIR"]

    return run_base_dir.joinpath(run_id[:2]).joinpath(run_id).resolve()


def resolve_content_path(run_id: str, key: RUN_DIR_STRUCTURE_KEYS) -> Path:
    run_dir: Path = resolve_run_dir_path(run_id)

    return run_dir.joinpath(RUN_DIR_STRUCTURE[key])


def read_state(run_id: str) -> State:
    try:
        with resolve_content_path(run_id, "state").open(mode="r") as f:
            state: State = f.readline().strip()  # type: ignore
            return state
    except Exception:
        return "UNKNOWN"


def count_system_state() -> Dict[State, int]:
    run_ids: List[str] = glob_all_run_ids()
    count: Dict[State, int] = dict(collections.Counter(
        [read_state(run_id) for run_id in run_ids]))

    return count


def glob_all_run_ids() -> List[str]:
    run_base_dir: Path = current_app.config["RUN_DIR"]
    run_requests: List[Path] = \
        list(run_base_dir.glob(f"**/{RUN_DIR_STRUCTURE['run_request']}"))
    run_ids: List[str] = \
        [run_request.parent.name for run_request in run_requests]

    return run_ids


def read_file(run_id: str, file_type: RUN_DIR_STRUCTURE_KEYS) -> Any:
    if "dir" in file_type:
        return None
    file_path = resolve_content_path(run_id, file_type)
    if file_path.exists() is False:
        return None
    with file_path.open(mode="r", encoding="utf-8") as f:
        content = f.read().strip()
        try:
            return json.loads(content)
        except json.JSONDecodeError:
            return content


def secure_filepath(filepath: str) -> Path:
    """
    We know `werkzeug.secure_filename()`.
    However, this function cannot represent the dir structure,

    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'

    Thus, it is incompatible with workflow engines such as snakemake.
    Therefore, We implemented this by referring to `werkzeug.secure_filename()`

    Please check `tests/unit_test/test_secure_filepath.py`

    Reference of `PurePath.parts`:
    >> > PurePath("/").parts
    ('/',)
    >> > PurePath("//").parts
    ('//',)
    >> > PurePath("/foo/bar").parts
    ('/', 'foo', 'bar')
    >> > PurePath("foo/bar").parts
    ('foo', 'bar')
    >> > PurePath("/foo/bar/").parts
    ('/', 'foo', 'bar')
    >> > PurePath("./foo/bar/").parts
    ('foo', 'bar')
    >> > PurePath("/../../foo/bar//").parts
    ('/', '..', '..', 'foo', 'bar')
    >> > PurePath("/../.../foo/bar//").parts
    ('/', '..', '...', 'foo', 'bar')
    """
    ascii_filepath = normalize("NFKD", filepath).encode(
        "ascii", "ignore").decode("ascii")
    pure_path = PurePath(ascii_filepath)
    nodes = []
    for node in pure_path.parts:
        # Change space to underscore
        node = "_".join(node.split())
        # Change [^A-Za-z0-9_.-] to empty.
        node = str(_filename_ascii_strip_re.sub("", node))
        if node not in ["", ".", ".."]:
            nodes.append(node)

    path = Path("/".join([str(node) for node in nodes]))

    return path


def write_file(run_id: str, key: RUN_DIR_STRUCTURE_KEYS, content: Any) -> None:
    file = resolve_content_path(run_id, key)
    file.parent.mkdir(parents=True, exist_ok=True)
    if file.suffix == ".json" and isinstance(content, (dict, list)):
        content = json.dumps(content, indent=2)
    with file.open(mode="w", encoding="utf-8") as f:
        f.write(str(content))


def sapporo_endpoint() -> str:
    host = request.host_url.strip("/")
    url_prefix = current_app.config['URL_PREFIX'].strip("/")
    endpoint = f"{host}/{url_prefix}".strip("/")

    return endpoint


def prepare_run_dir(run_id: str, run_request: RunRequest) -> None:
    run_dir = resolve_run_dir_path(run_id)
    run_dir.mkdir(parents=True, exist_ok=True)
    exe_dir = resolve_content_path(run_id, "exe_dir")
    exe_dir.mkdir(parents=True, exist_ok=True)
    exe_dir.chmod(0o777)
    outputs_dir: Path = resolve_content_path(run_id, "outputs_dir")
    outputs_dir.mkdir(parents=True, exist_ok=True)
    outputs_dir.chmod(0o777)

    write_file(run_id, "sapporo_config", dump_sapporo_config())
    write_file(run_id, "run_request", run_request)
    write_file(run_id, "wf_params", run_request["workflow_params"])
    write_file(run_id, "wf_engine_params",
               convert_wf_engine_params_str(run_request))

    write_file(run_id, "service_info", generate_service_info())
    write_file(run_id, "executable_workflows", generate_executable_workflows())
    with current_app.config["RUN_SH"].open(mode="r", encoding="utf-8") as f:
        run_sh_content = f.read()
    write_file(run_id, "run_sh", run_sh_content)

    yevis_metadata = request.form.get("yevis_metadata", None)
    if yevis_metadata is not None:
        write_file(run_id, "yevis_metadata", yevis_metadata)

    write_workflow_attachment(run_id, run_request)


def dump_sapporo_config() -> Dict[str, Any]:
    return {
        "sapporo_version": current_app.config["SAPPORO_VERSION"],
        "get_runs": current_app.config["GET_RUNS"],
        "workflow_attachment": current_app.config["WORKFLOW_ATTACHMENT"],
        "registered_only_mode": current_app.config["REGISTERED_ONLY_MODE"],
        "service_info": str(current_app.config["SERVICE_INFO"]),
        "executable_workflows": str(current_app.config["EXECUTABLE_WORKFLOWS"]),
        "run_sh": str(current_app.config["RUN_SH"]),
        "url_prefix": current_app.config["URL_PREFIX"],
        "sapporo_endpoint": sapporo_endpoint(),
    }


def convert_wf_engine_params_str(run_request: RunRequest) -> str:
    wf_engine_params: Optional[str] = run_request["workflow_engine_parameters"]
    params: List[str] = []
    if wf_engine_params is None:
        service_info = generate_service_info()
        default_wf_engine_dict = service_info["default_workflow_engine_parameters"]
        default_wf_engine_params = default_wf_engine_dict.get(
            run_request["workflow_engine_name"], [])
        for default_wf_engine_param in default_wf_engine_params:
            params.append(default_wf_engine_param.get("name", ""))
            params.append(default_wf_engine_param.get("default_value", ""))
    else:
        wf_engine_params_obj = json.loads(wf_engine_params)
        if isinstance(wf_engine_params_obj, list):
            params.extend(map(str, wf_engine_params_obj))
        elif isinstance(wf_engine_params_obj, dict):
            for key, value in wf_engine_params_obj.items():
                params.append(str(key))
                params.append(str(value))

    return " ".join(params)


def write_workflow_attachment(run_id: str, run_request: RunRequest) -> None:
    exe_dir = resolve_content_path(run_id, "exe_dir")
    if current_app.config["WORKFLOW_ATTACHMENT"]:
        workflow_attachment = request.files.getlist("workflow_attachment")
        for file_storage in workflow_attachment:
            if file_storage.filename:
                file_name = secure_filepath(file_storage.filename)
                file_path = exe_dir.joinpath(file_name).resolve()
                file_path.parent.mkdir(parents=True, exist_ok=True)
                file_storage.save(file_path)


# Called in run.sh
def download_workflow_attachment(inputted_run_dir: str) -> None:
    run_dir: Path = Path(inputted_run_dir).resolve()
    config_path = run_dir.joinpath(RUN_DIR_STRUCTURE["sapporo_config"])
    with config_path.open(mode="r", encoding="utf-8") as f:
        sapporo_config = json.load(f)
    endpoint = sapporo_config["sapporo_endpoint"]
    exe_dir = run_dir.joinpath(RUN_DIR_STRUCTURE["exe_dir"])
    run_request_path = run_dir.joinpath(RUN_DIR_STRUCTURE["run_request"])
    with run_request_path.open(mode="r", encoding="utf-8") as f:
        run_request = json.load(f)
    wf_attachment_obj: List[AttachedFile] = json.loads(run_request["workflow_attachment"] or "[]")
    for file in wf_attachment_obj:
        name = file["file_name"]
        url = file["file_url"]
        parsed_url = parse.urlparse(url)
        if parsed_url.scheme in ["http", "https"] and not url.startswith(endpoint):
            file_path = exe_dir.joinpath(secure_filepath(name)).resolve()
            file_path.parent.mkdir(parents=True, exist_ok=True)
            with requests.get(url, stream=True) as res:
                if res.status_code == 200:
                    with file_path.open(mode="wb") as f:
                        res.raw.decode_content = True
                        shutil.copyfileobj(res.raw, f, 1024 * 1024)


def fork_run(run_id: str) -> None:
    run_dir: Path = resolve_run_dir_path(run_id)
    stdout: Path = resolve_content_path(run_id, "stdout")
    stderr: Path = resolve_content_path(run_id, "stderr")
    cmd: str = f"/bin/bash {current_app.config['RUN_SH']} {run_dir}"
    write_file(run_id, "state", "QUEUED")
    with stdout.open(mode="w", encoding="utf-8") as f_stdout, \
            stderr.open(mode="w", encoding="utf-8") as f_stderr:
        process = Popen(shlex.split(cmd),  # pylint: disable=consider-using-with
                        cwd=str(run_dir),
                        env=os.environ.copy(),
                        encoding="utf-8",
                        stdout=f_stdout, stderr=f_stderr)
    pid: Optional[int] = process.pid
    if pid is not None:
        write_file(run_id, "pid", str(pid))


def cancel_run(run_id: str) -> None:
    state: State = read_state(run_id)
    if state == "RUNNING":
        write_file(run_id, "state", "CANCELING")
        pid: int = int(read_file(run_id, "pid"))
        os.kill(pid, signal.SIGUSR1)


def resolve_requested_file_path(run_id: str, subpath: str) -> Path:
    if Path(subpath).name[0] == ".":
        requested_path = secure_filepath(
            str(Path(subpath).parent)).joinpath(Path(subpath).name)
    else:
        requested_path = secure_filepath(subpath)
    run_dir_path = resolve_run_dir_path(run_id)

    return run_dir_path.joinpath(requested_path)


def path_hierarchy(original_path: Path, dir_path: Path) -> Any:
    hierarchy: Dict[str, Any] = {
        "type": "directory",
        "name": dir_path.name,
        "path": str(dir_path.relative_to(original_path)),
    }

    try:
        hierarchy["children"] = [
            path_hierarchy(original_path, dir_path.joinpath(child))
            for child in dir_path.iterdir()
        ]
    except Exception:
        hierarchy["type"] = "file"

    return hierarchy


# Called in run.sh
def dump_outputs_list(inputted_run_dir: str) -> None:
    run_dir: Path = Path(inputted_run_dir).resolve()
    run_id = run_dir.name
    config_path = run_dir.joinpath(RUN_DIR_STRUCTURE["sapporo_config"])
    with config_path.open(mode="r", encoding="utf-8") as f:
        sapporo_config = json.load(f)
    base_remote_url = \
        f"{sapporo_config['sapporo_endpoint']}/runs/{run_id}/data/"
    outdir_path: Path = run_dir.joinpath(RUN_DIR_STRUCTURE["outputs_dir"])
    output_files: List[Path] = sorted(list(walk_all_files(outdir_path)))
    outputs = []
    for output_file in output_files:
        outputs.append({
            "file_name": str(output_file.relative_to(outdir_path)),
            "file_url":
                f"{base_remote_url}{str(output_file.relative_to(run_dir))}"
        })
    output_path = run_dir.joinpath(RUN_DIR_STRUCTURE["outputs"])
    with output_path.open(mode="w", encoding="utf-8") as f:
        f.write(json.dumps(outputs, indent=2))


def walk_all_files(dir_path: Path) -> Iterable[Path]:
    for root, _, files in os.walk(dir_path):
        for file in files:
            yield Path(root).joinpath(file)



#!/usr/bin/env python3
# coding: utf-8
from pathlib import Path
from typing import Dict, Literal

SRC_DIR: Path = Path(__file__).parent.resolve()

DEFAULT_SERVICE_INFO: Path = SRC_DIR.joinpath("service-info.json").resolve()
DEFAULT_EXECUTABLE_WORKFLOWS: Path = \
    SRC_DIR.joinpath("executable_workflows.json").resolve()
DEFAULT_RUN_SH: Path = SRC_DIR.joinpath("run.sh").resolve()
DEFAULT_RUN_DIR = Path.cwd().joinpath("run").resolve()
DEFAULT_HOST: str = "127.0.0.1"
DEFAULT_PORT: int = 1122
DEFAULT_ACCESS_CONTROL_ALLOW_ORIGIN: str = "*"
DEFAULT_URL_PREFIX: str = "/"
GET_STATUS_CODE: int = 200
POST_STATUS_CODE: int = 200
DATE_FORMAT: str = "%Y-%m-%dT%H:%M:%S"

SERVICE_INFO_SCHEMA: Path = \
    SRC_DIR.joinpath("service-info.schema.json").resolve()
EXECUTABLE_WORKFLOWS_SCHEMA: Path = \
    SRC_DIR.joinpath("executable_workflows.schema.json").resolve()

RUN_DIR_STRUCTURE: Dict[str, str] = {
    "sapporo_config": "sapporo_config.json",
    "run_request": "run_request.json",
    "state": "state.txt",
    "exe_dir": "exe",
    "outputs_dir": "outputs",
    "outputs": "outputs.json",
    "wf_params": "exe/workflow_params.json",
    "start_time": "start_time.txt",
    "end_time": "end_time.txt",
    "exit_code": "exit_code.txt",
    "stdout": "stdout.log",
    "stderr": "stderr.log",
    "pid": "run.pid",
    "wf_engine_params": "workflow_engine_params.txt",
    "cmd": "cmd.txt",
    "task_logs": "task.log",
    "service_info": "service_info.json",
    "executable_workflows": "executable_workflows.json",
    "run_sh": "run.sh",
    "yevis_metadata": "yevis-metadata.yml",
}

RUN_DIR_STRUCTURE_KEYS = Literal[
    "sapporo_config",
    "run_request",
    "state",
    "exe_dir",
    "outputs_dir",
    "outputs",
    "wf_params",
    "start_time",
    "end_time",
    "exit_code",
    "stdout",
    "stderr",
    "pid",
    "wf_engine_params",
    "cmd",
    "task_logs",
    "service_info",
    "executable_workflows",
    "run_sh",
    "yevis_metadata",
]



#!/usr/bin/env python3
# coding: utf-8
from pathlib import Path

from setuptools import setup

BASE_DIR: Path = Path(__file__).parent.resolve()
LONG_DESCRIPTION: Path = BASE_DIR.joinpath("README.md")

setup(
    name="sapporo",
    version="1.4.6",
    description="The sapporo-service is a standard implementation conforming to "
    "the Global Alliance for Genomics and Health (GA4GH) Workflow Execution "
    "Service (WES) API specification.",
    long_description=LONG_DESCRIPTION.open(mode="r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="DDBJ(Bioinformatics and DDBJ Center)",
    author_email="t.ohta@nig.ac.jp",
    url="https://github.com/sapporo-wes/sapporo-service",
    license="Apache2.0",
    python_requires=">=3.7",
    platforms="any",
    packages=["sapporo", "sapporo.model"],
    package_data={
        "sapporo": [
            "executable_workflows.json",
            "executable_workflows.schema.json",
            "run.sh",
            "service-info.json",
            "service-info.schema.json",
            "ro-terms.csv",
        ]
    },
    include_package_data=True,
    install_requires=[
        "cwl-inputs-parser>=1.0.2",
        "flask-cors",
        "flask",
        "jsonschema",
        "psutil",
        "python-magic",
        "pyyaml",
        "requests",
        "rocrate",
        "uwsgi",
    ],
    tests_require=[
        "flake8",
        "isort",
        "mypy",
        "pytest",
        "types-PyYAML",
        "types-requests",
        "types-setuptools",
        "typing-extensions",
    ],
    extras_require={
        "tests": [
            "flake8",
            "isort",
            "mypy",
            "pytest",
            "types-PyYAML",
            "types-requests",
            "types-setuptools",
            "typing-extensions",
        ],
    },
    entry_points={
        "console_scripts": [
            "sapporo=sapporo.app:main",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Flask",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



#!/usr/bin/env python3
# coding: utf-8
from typing import Dict, List, Literal, Optional, TypedDict, Union

WorkflowTypes = Optional[Literal["CWL", "WDL", "SMK", "NFL"]]


class DefaultWorkflowEngineParameter(TypedDict):
    name: str
    type: str
    default_value: str


class Log(TypedDict):
    name: str
    cmd: List[str]
    start_time: Optional[str]
    end_time: Optional[str]
    stdout: Optional[str]
    stderr: Optional[str]
    exit_code: Optional[int]


class WorkflowTypeVersion(TypedDict):
    workflow_type_version: List[str]


State = Literal[
    "UNKNOWN",
    "QUEUED",
    "INITIALIZING",
    "RUNNING",
    "PAUSED",
    "COMPLETE",
    "EXECUTOR_ERROR",
    "SYSTEM_ERROR",
    "CANCELED",
    "CANCELING"
]


class ServiceInfo(TypedDict):
    workflow_type_versions: Dict[WorkflowTypes, WorkflowTypeVersion]
    supported_wes_versions: List[str]
    supported_filesystem_protocols: List[str]
    workflow_engine_versions: Dict[str, str]
    default_workflow_engine_parameters: Dict[str,
                                             List[DefaultWorkflowEngineParameter]]
    system_state_counts: Dict[State, int]
    auth_instructions_url: str
    contact_info_url: str
    tags: Dict[str, str]


class SecondaryFile(TypedDict):
    required: bool
    pattern: str


class WorkflowInput(TypedDict):
    default: Optional[Union[str, int, bool]]
    doc: Optional[str]
    id: str
    label: Optional[str]
    type: Literal["file", "directory", "any", "string", "integer", "boolean"]
    array: bool
    required: bool
    secondary_files: Optional[List[SecondaryFile]]


class ParseRequest(TypedDict):
    workflow_content: Optional[str]
    workflow_location: Optional[str]
    types_of_parsing: Optional[List[str]]


class ParseResult(TypedDict):
    workflow_type: Optional[WorkflowTypes]
    workflow_type_version: Optional[str]
    inputs: Optional[Union[List[WorkflowInput], str]]


class RunStatus(TypedDict):
    run_id: str
    state: State


class RunListResponse(TypedDict):
    runs: List[RunStatus]
    next_page_token: str


class AttachedFile(TypedDict):
    file_name: str
    file_url: str


class RunRequest(TypedDict):
    workflow_params: Optional[str]
    workflow_type: Optional[WorkflowTypes]
    workflow_type_version: Optional[str]
    tags: Optional[str]
    workflow_engine_name: str
    workflow_engine_parameters: Optional[str]
    workflow_url: Optional[str]
    workflow_name: Optional[str]
    workflow_attachment: Optional[str]


class RunLog(TypedDict):
    run_id: str
    request: RunRequest
    state: State
    run_log: Log
    task_logs: List[Log]
    outputs: List[AttachedFile]


class RunId(TypedDict):
    run_id: str


class ErrorResponse(TypedDict):
    msg: str
    status_code: int


class Workflow(TypedDict):
    workflow_name: str
    workflow_url: str
    workflow_type: str
    workflow_type_version: str
    workflow_attachment: List[AttachedFile]
